@@ -26,13 +26,9 @@
 #include <string.h>
 #include <ctype.h>
 #include <math.h>
-
-#ifdef HAVE_VALUES_H
-#include <values.h>
-#endif
-
 #include <time.h>
 #include <unistd.h>
+#include <langinfo.h>
 
 #include <gtk/gtk.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
@@ -192,7 +188,7 @@ static const gchar *conditions_str[24][13] = {
     /* SNOW          */ {N_("Snow"),                  "??",                                N_("Light snow"),          N_("Moderate snow"),          N_("Heavy snow"),          "??",                        "??",                           "??",                        N_("Snowstorm"),                N_("Blowing snowfall"),      N_("Snow showers"),             N_("Drifting snow"),          "??"                         },
     /* SNOW_GRAINS   */ {N_("Snow grains"),           "??",                                N_("Light snow grains"),   N_("Moderate snow grains"),   N_("Heavy snow grains"),   "??",                        "??",                           "??",                        "??",                           "??",                        "??",                           "??",                         "??"                         },
     /* ICE_CRYSTALS  */ {N_("Ice crystals"),          "??",                                "??",                      N_("Ice crystals"),           "??",                      "??",                        "??",                           "??",                        "??",                           "??",                        "??",                           "??",                         "??"                         },
-    /* ICE_PELLETS   */ {N_("Ice pellets"),           "??",                                N_("Few ice pellets"),     N_("Moderate ice pellets"),   N_("Heavy ice pellets"),   "??",                        "??",                           "??",                        N_("Ice pellet storm"),         "??",                        N_("Showers of ice pellets"),   "??",                         "??"                         },
+    /* ICE_PELLETS   */ {N_("Sleet"),           "??",                                N_("Little sleet"),     N_("Moderate sleet"),   N_("Heavy sleet"),   "??",                        "??",                           "??",                        N_("Sleet storm"),         "??",                        N_("Showers of sleet"),   "??",                         "??"                         },
     /* HAIL          */ {N_("Hail"),                  "??",                                "??",                      N_("Hail"),                   "??",                      "??",                        "??",                           "??",                        N_("Hailstorm"),                "??",                        N_("Hail showers"),             "??",                         "??",                        },
     /* SMALL_HAIL    */ {N_("Small hail"),            "??",                                "??",                      N_("Small hail"),             "??",                      "??",                        "??",                           "??",                        N_("Small hailstorm"),          "??",                        N_("Showers of small hail"),    "??",                         "??"                         },
     /* PRECIPITATION */ {N_("Unknown precipitation"), "??",                                "??",                      "??",                         "??",                      "??",                        "??",                           "??",                        "??",                           "??",                        "??",                           "??",                         "??"                         },
@@ -420,6 +416,97 @@ gweather_info_init (GWeatherInfo *info)
     gweather_info_reset (info);
 }
 
+static SoupCache *
+get_cache (void)
+{
+    SoupCache *cache;
+    char *filename;
+
+    filename = g_build_filename (g_get_user_cache_dir (),
+				 "libgweather", NULL);
+
+    if (g_mkdir_with_parents (filename, 0700) < 0) {
+	g_free (filename);
+	return NULL;
+    }
+
+    cache = soup_cache_new (filename, SOUP_CACHE_SINGLE_USER);
+
+    g_free (filename);
+    return cache;
+}
+
+static void
+dump_and_unref_cache (SoupCache *cache)
+{
+    soup_cache_dump (cache);
+    g_object_unref (cache);
+}
+
+static SoupSession *static_session;
+
+static SoupSession *
+ref_session (void)
+{
+    SoupSession *session;
+    SoupCache *cache;
+
+    session = static_session;
+
+    if (session != NULL)
+	return g_object_ref (session);
+
+    session = soup_session_async_new ();
+    soup_session_add_feature_by_type (session, SOUP_TYPE_PROXY_RESOLVER_DEFAULT);
+    soup_session_add_feature_by_type (session, SOUP_TYPE_CONTENT_DECODER);
+
+    cache = get_cache ();
+    soup_session_add_feature (session, SOUP_SESSION_FEATURE (cache));
+    g_object_set_data_full (G_OBJECT (session), "libgweather-cache", g_object_ref (cache),
+			    (GDestroyNotify) dump_and_unref_cache);
+
+    soup_cache_load (cache);
+    g_object_unref (cache);
+
+    static_session = session;
+    g_object_add_weak_pointer (G_OBJECT (session), (void**) &static_session);
+
+    return session;
+}
+
+/**
+ * gweather_info_store_cache:
+ *
+ * Ensures that any data cached from the network is stored to disk.
+ * Calling this is not necessary, as the cache will be saved when
+ * the last reference to a #GWeatherInfo will be dropped.
+ * On the other hand, it must be called if there is any chance that
+ * the application will be closed without unreffing all objects, such
+ * as when using a language binding that employs a GC.
+ */
+void
+gweather_info_store_cache (void)
+{
+    SoupCache *cache;
+
+    if (static_session == NULL)
+	return;
+
+    cache = g_object_get_data (G_OBJECT (static_session), "libgweather-cache");
+    soup_cache_dump (cache);
+}
+
+/**
+ * gweather_info_update:
+ * @info: a #GWeatherInfo
+ *
+ * Requests a reload of weather conditions and forecast data from
+ * enabled network services.
+ * This call does no synchronous IO: rather, the result is delivered
+ * by emitting the #GWeatherInfo::updated signal.
+ * Note that if no network services are enabled, the signal will not
+ * be emitted. See #GWeatherInfo:enabled-providers for details.
+ */
 void
 gweather_info_update (GWeatherInfo *info)
 {
@@ -432,10 +519,8 @@ gweather_info_update (GWeatherInfo *info)
 
     gweather_info_reset (info);
 
-    if (!priv->session) {
-	priv->session = soup_session_async_new ();
-	soup_session_add_feature_by_type (priv->session, SOUP_TYPE_PROXY_RESOLVER_DEFAULT);
-    }
+    if (!priv->session)
+	priv->session = ref_session ();
 
     if (priv->providers & GWEATHER_PROVIDER_METAR)
 	metar_start_open (info);
@@ -580,9 +665,39 @@ gweather_info_get_conditions (GWeatherInfo *info)
     return g_strdup(gweather_conditions_to_string (&info->priv->cond));
 }
 
+static gboolean
+is_locale_metric (void)
+{
+    const char *fmt;
+    fmt = nl_langinfo (_NL_MEASUREMENT_MEASUREMENT);
+
+    if (fmt && *fmt == 2)
+	return FALSE;
+    else
+	return TRUE;
+}
+
+static GWeatherTemperatureUnit
+temperature_unit_to_real (GWeatherTemperatureUnit unit)
+{
+    if (G_UNLIKELY (unit == GWEATHER_TEMP_UNIT_INVALID)) {
+	g_critical("Conversion to invalid temperature unit");
+	unit = GWEATHER_TEMP_UNIT_DEFAULT;
+    }
+
+    if (unit == GWEATHER_TEMP_UNIT_DEFAULT)
+	return is_locale_metric() ?
+	    GWEATHER_TEMP_UNIT_CENTIGRADE :
+	    GWEATHER_TEMP_UNIT_FAHRENHEIT;
+
+    return unit;
+}
+
 static gchar *
 temperature_string (gfloat temp_f, GWeatherTemperatureUnit to_unit, gboolean want_round)
 {
+    to_unit = temperature_unit_to_real (to_unit);
+
     switch (to_unit) {
     case GWEATHER_TEMP_UNIT_FAHRENHEIT:
 	if (!want_round) {
@@ -614,10 +729,10 @@ temperature_string (gfloat temp_f, GWeatherTemperatureUnit to_unit, gboolean wan
 
     case GWEATHER_TEMP_UNIT_INVALID:
     case GWEATHER_TEMP_UNIT_DEFAULT:
-    default:
-	g_critical ("Conversion to illegal temperature unit: %d", to_unit);
-	return g_strdup(C_("temperature unit", "Unknown"));
+	g_assert_not_reached ();
     }
+
+    return NULL;
 }
 
 gchar *
@@ -721,9 +836,27 @@ gweather_info_get_apparent (GWeatherInfo *info)
     return temperature_string (apparent, g_settings_get_enum (priv->settings, TEMPERATURE_UNIT), FALSE);
 }
 
+static GWeatherSpeedUnit
+speed_unit_to_real (GWeatherSpeedUnit unit)
+{
+    if (G_UNLIKELY (unit == GWEATHER_SPEED_UNIT_INVALID)) {
+	g_critical("Conversion to invalid speed unit");
+	unit = GWEATHER_SPEED_UNIT_DEFAULT;
+    }
+
+    if (unit == GWEATHER_SPEED_UNIT_DEFAULT)
+	return is_locale_metric() ?
+	    GWEATHER_SPEED_UNIT_KPH :
+	    GWEATHER_SPEED_UNIT_KNOTS;
+
+    return unit;
+}
+
 static gchar *
 windspeed_string (gfloat knots, GWeatherSpeedUnit to_unit)
 {
+    to_unit = speed_unit_to_real (to_unit);
+
     switch (to_unit) {
     case GWEATHER_SPEED_UNIT_KNOTS:
 	/* TRANSLATOR: This is the wind speed in knots */
@@ -744,9 +877,7 @@ windspeed_string (gfloat knots, GWeatherSpeedUnit to_unit)
 	return g_strdup_printf(_("Beaufort force %.1f"), WINDSPEED_KNOTS_TO_BFT (knots));
     case GWEATHER_SPEED_UNIT_INVALID:
     case GWEATHER_SPEED_UNIT_DEFAULT:
-    default:
-	g_critical ("Conversion to illegal speed unit: %d", to_unit);
-	return g_strdup(C_("speed unit", "Unknown"));
+	g_assert_not_reached ();
     }
 
     return NULL;
@@ -781,10 +912,27 @@ gweather_info_get_wind (GWeatherInfo *info)
     }
 }
 
+static GWeatherPressureUnit
+pressure_unit_to_real (GWeatherPressureUnit unit)
+{
+    if (G_UNLIKELY (unit == GWEATHER_PRESSURE_UNIT_INVALID)) {
+	g_critical("Conversion to invalid pressure unit");
+	unit = GWEATHER_PRESSURE_UNIT_DEFAULT;
+    }
+
+    if (unit == GWEATHER_PRESSURE_UNIT_DEFAULT)
+	return is_locale_metric() ?
+	    GWEATHER_PRESSURE_UNIT_MM_HG :
+	    GWEATHER_PRESSURE_UNIT_INCH_HG;
+
+    return unit;
+}
+
 gchar *
 gweather_info_get_pressure (GWeatherInfo *info)
 {
     GWeatherInfoPrivate *priv;
+    GWeatherPressureUnit unit;
 
     g_return_val_if_fail (GWEATHER_IS_INFO (info), NULL);
 
@@ -795,7 +943,8 @@ gweather_info_get_pressure (GWeatherInfo *info)
     if (priv->pressure < 0.0)
         return g_strdup(C_("pressure", "Unknown"));
 
-    switch (g_settings_get_enum (priv->settings, PRESSURE_UNIT)) {
+    unit = pressure_unit_to_real (g_settings_get_enum (priv->settings, PRESSURE_UNIT));
+    switch (unit) {
     case GWEATHER_PRESSURE_UNIT_INCH_HG:
 	/* TRANSLATOR: This is pressure in inches of mercury */
 	return g_strdup_printf(_("%.2f inHg"), priv->pressure);
@@ -817,18 +966,33 @@ gweather_info_get_pressure (GWeatherInfo *info)
 
     case GWEATHER_PRESSURE_UNIT_INVALID:
     case GWEATHER_PRESSURE_UNIT_DEFAULT:
-    default:
-	g_critical ("Conversion to illegal pressure unit");
-	return g_strdup(C_("pressure unit", "Unknown"));
+	g_assert_not_reached ();
     }
 
     return NULL;
 }
 
+static GWeatherDistanceUnit
+distance_unit_to_real (GWeatherDistanceUnit unit)
+{
+    if (G_UNLIKELY (unit == GWEATHER_DISTANCE_UNIT_INVALID)) {
+	g_critical("Conversion to invalid distance unit");
+	unit = GWEATHER_DISTANCE_UNIT_DEFAULT;
+    }
+
+    if (unit == GWEATHER_DISTANCE_UNIT_DEFAULT)
+	return is_locale_metric() ?
+	    GWEATHER_DISTANCE_UNIT_METERS :
+	    GWEATHER_DISTANCE_UNIT_MILES;
+
+    return unit;
+}
+
 gchar *
 gweather_info_get_visibility (GWeatherInfo *info)
 {
     GWeatherInfoPrivate *priv;
+    GWeatherDistanceUnit unit;
 
     g_return_val_if_fail (GWEATHER_IS_INFO (info), NULL);
     priv = info->priv;
@@ -838,7 +1002,8 @@ gweather_info_get_visibility (GWeatherInfo *info)
     if (priv->visibility < 0.0)
         return g_strdup (C_("visibility", "Unknown"));
 
-    switch (g_settings_get_enum (priv->settings, DISTANCE_UNIT)) {
+    unit = distance_unit_to_real (g_settings_get_enum (priv->settings, DISTANCE_UNIT));
+    switch (unit) {
     case GWEATHER_DISTANCE_UNIT_MILES:
 	/* TRANSLATOR: This is the visibility in miles */
 	return g_strdup_printf (_("%.1f miles"), priv->visibility);
@@ -851,9 +1016,7 @@ gweather_info_get_visibility (GWeatherInfo *info)
 
     case GWEATHER_DISTANCE_UNIT_INVALID:
     case GWEATHER_DISTANCE_UNIT_DEFAULT:
-    default:
-	g_critical ("Conversion to illegal visibility unit");
-	return g_strdup (C_("visibility unit", "Unknown"));
+	g_assert_not_reached ();
     }
 
     return NULL;
@@ -1020,13 +1183,13 @@ gweather_info_get_weather_summary (GWeatherInfo *info)
  * gweather_info_is_daytime:
  * @info: a #GWeatherInfo
  *
- * Returns whether it is daytime (that is, if the sun is visible)
- * or not at the location and the point of time referred by @info.
- * This is mostly equivalent to comparing the return value
- * of gweather_info_get_value_sunrise() and
- * gweather_info_get_value_sunset(), but it accounts also
- * for midnight sun and polar night, for locations within
- * the Artic and Antartic circles.
+ * Returns: Whether it is daytime (that is, if the sun is visible)
+ *   or not at the location and the point of time referred by @info.
+ *   This is mostly equivalent to comparing the return value
+ *   of gweather_info_get_value_sunrise() and
+ *   gweather_info_get_value_sunset(), but it accounts also
+ *   for midnight sun and polar night, for locations within
+ *   the Artic and Antartic circles.
  */
 gboolean
 gweather_info_is_daytime (GWeatherInfo *info)
@@ -1180,6 +1343,95 @@ gweather_info_get_icon_name (GWeatherInfo *info)
     return icon_buffer;
 }
 
+const gchar *
+gweather_info_get_symbolic_icon_name (GWeatherInfo *info)
+{
+    GWeatherInfoPrivate *priv;
+    GWeatherConditions   cond;
+    GWeatherSky          sky;
+    gboolean             daytime;
+
+    g_return_val_if_fail (GWEATHER_IS_INFO (info), NULL);
+
+    priv = info->priv;
+
+    _gweather_info_ensure_sun (info);
+    _gweather_info_ensure_moon (info);
+
+    cond = priv->cond;
+    sky = priv->sky;
+
+    if (cond.significant) {
+	if (cond.phenomenon != GWEATHER_PHENOMENON_NONE &&
+	    cond.qualifier == GWEATHER_QUALIFIER_THUNDERSTORM)
+            return "weather-storm-symbolic";
+
+        switch (cond.phenomenon) {
+	case GWEATHER_PHENOMENON_INVALID:
+	case GWEATHER_PHENOMENON_LAST:
+	case GWEATHER_PHENOMENON_NONE:
+	    break;
+
+	case GWEATHER_PHENOMENON_DRIZZLE:
+	case GWEATHER_PHENOMENON_RAIN:
+	case GWEATHER_PHENOMENON_UNKNOWN_PRECIPITATION:
+	case GWEATHER_PHENOMENON_HAIL:
+	case GWEATHER_PHENOMENON_SMALL_HAIL:
+	    return "weather-showers-symbolic";
+
+	case GWEATHER_PHENOMENON_SNOW:
+	case GWEATHER_PHENOMENON_SNOW_GRAINS:
+	case GWEATHER_PHENOMENON_ICE_PELLETS:
+	case GWEATHER_PHENOMENON_ICE_CRYSTALS:
+	    return "weather-snow-symbolic";
+
+	case GWEATHER_PHENOMENON_TORNADO:
+	case GWEATHER_PHENOMENON_SQUALL:
+	    return "weather-storm-symbolic";
+
+	case GWEATHER_PHENOMENON_MIST:
+	case GWEATHER_PHENOMENON_FOG:
+	case GWEATHER_PHENOMENON_SMOKE:
+	case GWEATHER_PHENOMENON_VOLCANIC_ASH:
+	case GWEATHER_PHENOMENON_SAND:
+	case GWEATHER_PHENOMENON_HAZE:
+	case GWEATHER_PHENOMENON_SPRAY:
+	case GWEATHER_PHENOMENON_DUST:
+	case GWEATHER_PHENOMENON_SANDSTORM:
+	case GWEATHER_PHENOMENON_DUSTSTORM:
+	case GWEATHER_PHENOMENON_FUNNEL_CLOUD:
+	case GWEATHER_PHENOMENON_DUST_WHIRLS:
+	    return "weather-fog-symbolic";
+        }
+    }
+
+    daytime = gweather_info_is_daytime (info);
+
+    switch (sky) {
+    case GWEATHER_SKY_INVALID:
+    case GWEATHER_SKY_LAST:
+    case GWEATHER_SKY_CLEAR:
+	if (daytime)
+	    return "weather-clear-symbolic";
+	else
+	    return "weather-clear-night-symbolic";
+
+    case GWEATHER_SKY_BROKEN:
+    case GWEATHER_SKY_SCATTERED:
+    case GWEATHER_SKY_FEW:
+	if (daytime)
+	    return "weather-few-clouds-symbolic";
+	else
+	    return "weather-few-clouds-night-symbolic";
+
+    case GWEATHER_SKY_OVERCAST:
+	return "weather-overcast-symbolic";
+
+    default: /* unrecognized */
+	return NULL;
+    }
+}
+
 static gboolean
 temperature_value (gdouble temp_f,
 		   GWeatherTemperatureUnit to_unit,
@@ -1194,6 +1446,7 @@ temperature_value (gdouble temp_f,
 
     if (to_unit == GWEATHER_TEMP_UNIT_DEFAULT)
 	    to_unit = g_settings_get_enum (settings, TEMPERATURE_UNIT);
+    to_unit = temperature_unit_to_real (to_unit);
 
     switch (to_unit) {
         case GWEATHER_TEMP_UNIT_FAHRENHEIT:
@@ -1207,9 +1460,7 @@ temperature_value (gdouble temp_f,
 	    break;
         case GWEATHER_TEMP_UNIT_INVALID:
         case GWEATHER_TEMP_UNIT_DEFAULT:
-	default:
-	    ok = FALSE;
-	    break;
+	    g_assert_not_reached ();
     }
 
     return ok;
@@ -1230,6 +1481,7 @@ speed_value (gdouble            knots,
 
     if (to_unit == GWEATHER_SPEED_UNIT_DEFAULT)
 	    to_unit = g_settings_get_enum (settings, SPEED_UNIT);
+    to_unit = speed_unit_to_real (to_unit);
 
     switch (to_unit) {
         case GWEATHER_SPEED_UNIT_KNOTS:
@@ -1249,9 +1501,7 @@ speed_value (gdouble            knots,
 	    break;
         case GWEATHER_SPEED_UNIT_INVALID:
         case GWEATHER_SPEED_UNIT_DEFAULT:
-        default:
-            ok = FALSE;
-            break;
+	    g_assert_not_reached ();
     }
 
     return ok;
@@ -1272,6 +1522,7 @@ pressure_value (gdouble               inHg,
 
     if (to_unit == GWEATHER_PRESSURE_UNIT_DEFAULT)
 	    to_unit = g_settings_get_enum (settings, PRESSURE_UNIT);
+    to_unit = pressure_unit_to_real (to_unit);
 
     switch (to_unit) {
         case GWEATHER_PRESSURE_UNIT_INCH_HG:
@@ -1294,9 +1545,7 @@ pressure_value (gdouble               inHg,
 	    break;
         case GWEATHER_PRESSURE_UNIT_INVALID:
         case GWEATHER_PRESSURE_UNIT_DEFAULT:
-        default:
-	    ok = FALSE;
-	    break;
+	    g_assert_not_reached ();
     }
 
     return ok;
@@ -1317,6 +1566,7 @@ distance_value (gdouble               miles,
 
     if (to_unit == GWEATHER_DISTANCE_UNIT_DEFAULT)
 	    to_unit = g_settings_get_enum (settings, DISTANCE_UNIT);
+    to_unit = distance_unit_to_real (to_unit);
 
     switch (to_unit) {
         case GWEATHER_DISTANCE_UNIT_MILES:
@@ -1330,9 +1580,7 @@ distance_value (gdouble               miles,
             break;
         case GWEATHER_DISTANCE_UNIT_INVALID:
         case GWEATHER_DISTANCE_UNIT_DEFAULT:
-        default:
-	    ok = FALSE;
-	    break;
+	    g_assert_not_reached ();
     }
 
     return ok;
@@ -1876,6 +2124,14 @@ gweather_info_class_init (GWeatherInfoClass *klass)
 				G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);
     g_object_class_install_property (gobject_class, PROP_ENABLED_PROVIDERS, pspec);
 
+    /**
+     * GWeatherInfo::updated:
+     * @object: the emitter of the signal.
+     *
+     * This signal is emitted after the initial fetch of the weather
+     * data from upstream services, and after every successful call
+     * to @gweather_info_update().
+     */
     gweather_info_signals[SIGNAL_UPDATED] = g_signal_new ("updated",
 							  GWEATHER_TYPE_INFO,
 							  G_SIGNAL_RUN_FIRST,
@@ -1892,8 +2148,10 @@ gweather_info_class_init (GWeatherInfoClass *klass)
  * @forecast_type: the type of forecast requested
  *
  * Builds a new #GWeatherInfo that will provide weather information about
- * @location. The returned info will not be ready until the #GWeatherInfo::updated signal
- * is emitted.
+ * @location.
+ * Note that, as compared to g_object_new(), this will call gweather_info_update()
+ * on the resulting info, which will not be ready until the #GWeatherInfo::updated
+ * signal is emitted.
  *
  * Returns: (transfer full): a new #GWeatherInfo
  */
